import uos as os


def walk_directories(initial_directory="/"): # pragma: no cover
    if not initial_directory.startswith("/"):
        initial_directory = f"/{initial_directory}"
    directories = [initial_directory]
    paths = {}
    while directories:
        for directory in directories:
            new_directories, files = recursive_walk(directory)
            paths[directory] = files
            directories = new_directories
    return paths

def recursive_walk(initial_directory): # pragma: no cover
    files = []
    directories = []
    unprocessed_directory = os.ilistdir(initial_directory)
    if unprocessed_directory:
        for item in unprocessed_directory:
            if item[1] == 16384:
                if initial_directory == "/":
                    initial_directory = ""
                directories.append(f"{initial_directory}/{item[0]}")
            else:
                files.append(item[0])
    return (directories, files)

def unquote(string):
    bytes_string = string.encode("utf-8")
    split_byte_string = bytes_string.split(b'%')
    if len(split_byte_string) == 1:
        return bytes_string
    final_byte_string = [split_byte_string[0]]
    for split in split_byte_string[1:]:
        final_byte_string.append(bytes([int(split[:2], 16)]))
        final_byte_string.append(split[2:])
    return b"".join(final_byte_string)



from picowebrouter.utils import unquote
class Request:
    def __init__(self,
        raw=None,
        method=None,
        path=None,
        query_strings=None,
        protocol=None,
        headers=None,
        body=None
    ):
        self.raw = raw
        self.method = method
        self.path = path
        self.query_strings = query_strings
        self.protocol = protocol
        self.headers = headers
        self.body = body

    
    @classmethod
    def parse_request(cls, request):
        body = None
        try:
            request, body = request.split("\r\n\r\n")
        except ValueError:
            request = request
        split_request = request.split("\r\n")
        request_line = split_request.pop(0)
        method, path, protocol = request_line.split(" ")
        query_strings = {}
        if "?" in path:
            path, raw_query = path.split("?")
            raw_query = raw_query.split("&")
            for query in raw_query:
                key, value = query.split("=")
                query_strings[key] = unquote(value).decode("utf-8")
            query_strings = query_strings

        headers = {}
        for header in split_request:
            key, value = header.split(": ")
            headers[key] = value
        headers = headers
        return cls(
            request,
            method,
            path,
            query_strings,
            protocol,
            headers,
            body
        )



from .objects.request import *
from .utils import walk_directories
import socket
import uos as os


class WebRouter:
    def __init__(self, ip, port, default, static_location="static"):
        self.routes = {"default": default}
        self.socket = self.open_socket(ip, port)
        self.static_location = static_location
        self.static_files = []
        self.add_static()

    def route(self, path):
        print(path)
        def router(function):
            def wrapper(*args, **kwargs):
                self.routes[path] = function

            return wrapper

        return router

    def add_static(self):
        static_files = walk_directories(self.static_location)
        if static_files:
            for directory in static_files.keys():
                for file in static_files[directory]:
                    self.static_files.append(f"{directory}/{file}")

    @staticmethod
    def open_socket(ip, port):
        address = (ip, port)
        new_socket = socket.socket()
        new_socket.bind((ip, port))
        new_socket.listen(1)
        return new_socket

    @staticmethod
    def determine_mimetype(path):
        mimetype = "text/html"
        if path.endswith(".js"):
            mimetype = "text/javascript"
        if path.endswith(".css"):
            mimetype = "text/css"
        return mimetype

    def serve(self): # pragma: no cover
        while True:
            client = self.socket.accept()[0]
            request = client.recv(1024)
            if request:
                request = Request.parse_request(request.decode("utf-8"))
                path = request.path
                header = "HTTP/1.1 200 OK\n"
                mimetype = self.determine_mimetype(path)
                if path in self.routes.keys():
                    webpage = self.routes[path](request=request)
                elif path in self.static_files:
                    with open(path, "rb") as static:
                        webpage = static.read()
                else:
                    webpage = self.routes["default"](request=request)
                header += f"Content-Type: {mimetype}\n\n"
                client.sendall(webpage)
                client.close()

